namespace StudySystemClient.Response {
	public class Response {
		public int transaction_id { get; private set; }
		public Body body { get; private set; }

		public Response.from_bytes(uint8[] bytes) throws DecodeError {
			Der.Sequence sequence;
			try {
				sequence = Der.decode(bytes) as Der.Sequence;
			} catch (Der.DecodeError e) {
				throw new DecodeError.INVALID_RESPONSE(
					"Response was not valid DER: " + e.message);
			}
			if (sequence == null) {
				throw new DecodeError.INVALID_RESPONSE(
					"Response was not a SEQUENCE");
			}
			if (sequence.value.length < 2) {
				throw new DecodeError.INVALID_RESPONSE(
					"Too few fields in Response: %u (expected 2)",
					sequence.value.length);
			}
			var id_datum = sequence.value[0] as Der.Integer;
			if (id_datum == null) {
				throw new DecodeError.INVALID_RESPONSE(
					"Response transactionId was not an INTEGER");
			}
			transaction_id = (int)id_datum.value;
			body = Body.from_datum(sequence.value[1]);
		}
	}

	public errordomain DecodeError {
		INVALID_BODY,
		INVALID_RESPONSE,
		NOT_IMPLEMENTED,
	}

	public abstract class Body {
		protected enum Tag {
			ERROR = 0,
			ACK = 1,
			ACTIVITIES = 2,
		}

		internal static Body from_datum(Der.Datum datum) throws DecodeError {
			var choice = datum as Der.Choice;
			if (choice == null) {
				throw new DecodeError.INVALID_BODY(
					"ResponseBody was not a CHOICE");
			}
			switch (choice.id) {
			case Tag.ERROR:
				return new Error.from_datum(choice.value);
			case Tag.ACK:
				return new Ack.from_datum(choice.value);
			case Tag.ACTIVITIES:
				return new Activities.from_datum(choice.value);
			default:
				throw new DecodeError.INVALID_BODY(
					"Invalid ResponseBody tag");
			}
		}
	}

	public class Error : Body {
		public enum Value {
			INVALID_REQUEST = 0,
			INVALID_ARGUMENTS = 1,
			SERVER_ERROR = 2,

			COUNT;

			public string to_string() {
				switch (this) {
				case INVALID_REQUEST:
					return "Invalid request";
				case INVALID_ARGUMENTS:
					return "Invalid arguments";
				case SERVER_ERROR:
					return "Server error";
				default:
					return "Unknown error";
				}
			}
		}

		public Value value { get; private set; }

		internal Error.from_datum(Der.Datum datum) throws DecodeError {
			var enumerated = datum as Der.Enumerated;
			if (enumerated == null) {
				throw new DecodeError.INVALID_BODY(
					"Error was not an ENUMERATED");
			}
			if (enumerated.value < 0 || enumerated.value >= Value.COUNT) {
				throw new DecodeError.INVALID_BODY(
					"Error type was not in range 0..$(Value.COUNT)");
			}
			value = (Value)enumerated.value;
		}
	}

	public class Ack : Body {
		internal Ack.from_datum(Der.Datum datum) throws DecodeError {
			var @null = datum as Der.Null;
			if (@null == null) {
				throw new DecodeError.INVALID_BODY("Ack was not NULL");
			}
		}
	}

	public class Activities : Body {
		public Array<Activity> value { get; private set; }

		internal Activities.from_datum(Der.Datum datum) throws DecodeError {
			value = new Array<Activity>();
			if (datum is Der.Sequence) {
				foreach (var activity_datum in datum.value)
					value.append_val(activity_from_datum(activity_datum));
			} else {
				throw new DecodeError.INVALID_BODY(
					"Activities was not a SEQUENCE");
			}
		}

		private static Activity activity_from_datum(Der.Datum datum)
			throws DecodeError {
			if (datum is Der.Sequence) {
				var fields = datum.value;
				if (fields.length < 3) {
					throw new DecodeError.INVALID_BODY(
						"Too few fields in Activity: %u (expected 3)",
						fields.length);
				}
				var subject = get_string("Activity.subject", fields[0]);
				var activity_type
					= get_activity_type("Activity.type", fields[1]);
				var priority_int = get_int("Activity.priority", fields[2]);
				var priority = (double)priority_int / 100.0;
				return { subject, activity_type, priority };
			} else {
				throw new DecodeError.INVALID_BODY(
					"Activity was not a SEQUENCE");
			}
		}

		private static int get_int(string name, Der.Datum datum)
			throws DecodeError {
			if (datum is Der.Integer)
				return (int)datum.value;
			throw new DecodeError.INVALID_BODY(@"$name was not an INTEGER");
		}

		private static string get_string(string name, Der.Datum datum)
			throws DecodeError {
			if (datum is Der.Utf8String)
				return datum.value;
			throw new DecodeError.INVALID_BODY(
				@"$name was not a UTF8String");
		}

		private static ActivityType get_activity_type(
			string name, Der.Datum datum) throws DecodeError {
			if (datum is Der.Enumerated) {
				var value = datum.value;
				if (0 <= value < ActivityType.COUNT)
					return (ActivityType)value;
				throw new DecodeError.INVALID_BODY(
					"Invalid value for ActivityType: %lld", value);
			}
			throw new DecodeError.INVALID_BODY(
				@"$name was not an ENUMERATED");
		}
	}
}
